# -*- coding: utf-8 -*-

# This file is part of Youba (Youpi back end)
# Copyright (C) 2011 AlphaK - http://www.alphak.net/
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging, sys, os, settings
from datetime import datetime, timedelta
import time
import threading
from classes import CommandPart, Action, WaitMode, Motor, Rotation, Unit
from classes import SpecialAction, InstructionChainModifier, InstructionChain
from parallelinterface import ParallelInterface
log = logging.getLogger('backend')
log.addHandler(logging.StreamHandler())


class Backend(threading.Thread):
	
	def __init__(self, name = '', instructionChain = None, lock=None):
		threading.Thread.__init__(self)
		self.name = name
		self.instructionChain = instructionChain
		self.lock = lock
		self._stopevent = threading.Event()
		self.currentWordRotation = None
		self.dumpfile = None
		self.dumpbuffer = ''
		self.p = None
		self.isDebug = False
		
		if settings.LOG_LEVEL == 'debug':
			self.isDebug = True
		
		try:
			self.p = ParallelInterface()
		except Exception,e:
			# We can't do much
			raise e
		
		try:
			self.dumpfile = open('%s/%s' % (os.path.dirname(os.path.abspath(__file__)), settings.DUMP_FILE), 'w')
		except Exception,e:
			# We can't do much
			raise e			
		
		self.SLEEP_RESOLUTION = self.searchForBestSleepResolution()
		self.resetCounters()
		self.initRobot()
		
		
	def stop(self):
		if self.dumpfile != None:
			try:
				self.dumpfile.close()
			except Exception,e:
				pass
		self._stopevent.set()
		
				
	def run(self):
		"""
		Backend core function, deals with the main job.
		"""
		while not self._stopevent.isSet():
			
			if(len(self.instructionChain) == 0):
				self._stopevent.wait(0.05)
				
			else:			
				# Acquire lock, during the time needed to make all motors move one step
				self.lock.acquire()
				
				# InstructionSets may suddenly change in a threaded environment, so double check after lock
				if(len(self.instructionChain) > 0):
				
					# Only deal with first instructionSet
					firstInstructionSet = self.instructionChain.instructionSet[0]
					
					# First loop: move arms
					for (motor, instruction) in firstInstructionSet.instructions.iteritems():
						now = datetime.now()
						if now >= instruction.timeNextMotorCommand and not instruction.isTerminated:
													
							# Deal with the time limit
							if instruction.timeLimit != None and now >= instruction.timeLimit:
								instruction.isTerminated = True
							
							# If the instruction is still alive	
							if not instruction.isTerminated:
									
								# In case of a new instruction:
								if(instruction.isNew):
									
									# Init backend words if Rotation.INITIAL
									if instruction.rotation == Rotation.INITIAL:
										counter = self.instructionChain.counters[motor]
										if counter > 0:
											instruction.rotation = Rotation.LEFT
										else:
											instruction.rotation = Rotation.RIGHT
										instruction.unit = Unit.HALFSTEP
										instruction.value = abs(counter)
										instruction.halfStepLimit = abs(counter)
										if instruction.value == 0:
											instruction.isMoveRequired = False
										instruction.setBackendWords()
									
									# Writes rotation byte to parallel port if instruction requests a change
									if self.currentWordRotation == None:
										self.currentWordRotation = 0x80
									if instruction.wordRotationMask != None:
										if self.currentWordRotation & instruction.wordRotationMask != Rotation.MAP_BITWISE_ROTATIONMASK[instruction.rotation] << Motor.MAP_BITWISE_HALFSTEP_ID[motor]:
											newWordRotation = self.currentWordRotation ^ (1 << Motor.MAP_BITWISE_HALFSTEP_ID[motor])
											byteSequence = [newWordRotation, newWordRotation & 0x3F]
											for byte in byteSequence:
												self.write(byte)
												self.appendBuffer(byte)
											self.currentWordRotation = newWordRotation
												
								# Only when the instruction requires a move
								if instruction.isMoveRequired:
									
									# Writes step impulse to parallel port
									byteSequence = [instruction.wordMotorCommand1, instruction.wordMotorCommand2]
									for byte in byteSequence:
										self.write(byte)
										if instruction.isNew or self.isDebug:
											self.appendBuffer(byte)
									
									# Increment step counters...
									stepToAdd = 1
									if instruction.rotation == Rotation.LEFT:
										stepToAdd = -1
									self.instructionChain.counters[motor] = self.instructionChain.counters[motor] + stepToAdd
									# ... modulo -360°/+360°
									if self.instructionChain.counters[motor] >= Motor.MAP_MAX_ROTATION_STEPS[motor]:
										self.instructionChain.counters[motor] = self.instructionChain.counters[motor] - Motor.MAP_MAX_ROTATION_STEPS[motor]
									elif self.instructionChain.counters[motor] <= -Motor.MAP_MAX_ROTATION_STEPS[motor]:
										self.instructionChain.counters[motor] = self.instructionChain.counters[motor] + Motor.MAP_MAX_ROTATION_STEPS[motor]
									
								# Flag the instruction as 'not new anymore'
								if instruction.isNew:
									instruction.isNew = False
									
								# Sets the cooldown time
								if instruction.motorCommandInterval != None:
									instruction.timeNextMotorCommand = datetime.now() + timedelta(microseconds=instruction.motorCommandInterval)
								
								# Deal with the halfstep limit
								instruction.halfStepCounter = instruction.halfStepCounter + 1
								if instruction.halfStepLimit != None and instruction.halfStepCounter >= instruction.halfStepLimit:
									instruction.isTerminated = True
						
					# Second loop: purge instructionSet, depending on WaitMode
					isAllInstrutionsTerminated = None
					if firstInstructionSet.waitMode == WaitMode.WAIT:
						# The backend waits the end of all instructions of the instructionSet
						isAllInstrutionsTerminated = True
						for (motor, instruction) in firstInstructionSet.instructions.iteritems():
							if instruction.isTerminated == False:
								isAllInstrutionsTerminated = False
								break
					else:
						# The backend considers the instructionSet as over as soon as the first instruction has ended
						isAllInstrutionsTerminated = False
						for (motor, instruction) in firstInstructionSet.instructions.iteritems():
							if instruction.isTerminated == True:
								isAllInstrutionsTerminated = True
								break
						
					if isAllInstrutionsTerminated:
						self.instructionChain.pop()
						self.dump()
					
				# Finally release lock
				self.lock.release()
				
				# And sleep the required amount of time
				self.hiresSleep(self.SLEEP_RESOLUTION)
				
			
		log.info('Thread [%s] terminated correctly' % self.name)


	def hiresSleep(self, microsec):
		"""
		Sleeps for the amount of microseconds given as parameter.
		Depending on the system and the OS, this function may sleep more than expected.
		"""
		start = datetime.now()
		end = start + timedelta(microseconds=int(microsec))
		while datetime.now() < end:
			time.sleep(0)
		end = datetime.now()
		delta = end-start
		return delta.seconds*1000000 + delta.microseconds


	def searchForBestSleepResolution(self):
		"""
		Searches the best system sleep resolution between 500 & 1500µs
		
		Optimal sleep resolution to run the robot to its maximal speed is normally 500µs as it is the GCD of 1500 and 2000.
		However, lots of OSes are not able to sleep with an accurate microsecond resolution.
		So this function tries to find the sleep resolution that will be optimal with the current system and OS.
		In order to run the robot at maximal speed, run this program on a real-time kernel OS.
		"""
		
		iterations = 10
		listOfMicrosecValuesToTest = range(500, 1600, 100)
		resultTable=[]
		
		for microsec in listOfMicrosecValuesToTest:
			averageError = sum(abs(self.hiresSleep(microsec) - microsec) for i in xrange(iterations)) / iterations
			drift = float(averageError)/float(microsec)
			log.debug('Try for sleep resolution of %dµs: Average error = %dµs, Drift = %0.3f' % (microsec, averageError, drift))
			ustuple=(drift, microsec, averageError)
			resultTable.append(ustuple)
		
		# Takes smallest drift, then smallest microsec value
		resultTable.sort()
		(drift, microsec, averageError) = resultTable[0]
		log.debug('System optimal sleep resolution is %dµs (Average error = %dµs, Drift = %0.3f)' % (microsec, averageError, drift))
		if(microsec != 500):
			log.warning('System optimal sleep resolution is %dµs. Robot maximal speed is not guaranteed. Try to use a real-time kernel for better results.' % microsec)
		return microsec


	def initRobot(self):
		"""
		Initializes the robot microcontroler interface, done each time the program is started.
		"""
		# Empty instruction chain
		self.lock.acquire()
		self.instructionChain.empty()
		self.lock.release()
		
		# Initialize rotation values
		self.currentWordRotation = 0x80
		
		# Send the init sequence
		byteSequence = [0x47, 0x00, 0x80, 0x00]
		for byte in byteSequence:
			self.write(byte)
			self.appendBuffer(byte)

		# Flush dump file
		self.dump()
		
	
	def resetCounters(self):
		"""
		Reset counters for each motor
		"""
		self.lock.acquire()
		self.instructionChain.resetCounters()
		self.lock.release()
		
		
	def readCounters(self):
		"""
		Read counters for each motor
		"""
		self.lock.acquire()
		c = [
			self.instructionChain.counters[Motor.BASE],
			self.instructionChain.counters[Motor.SHOULDER],
			self.instructionChain.counters[Motor.ELBOW],
			self.instructionChain.counters[Motor.WRIST],
			self.instructionChain.counters[Motor.HAND],
			self.instructionChain.counters[Motor.PLIER]
		]
		self.lock.release()
		return c
	
	
	def appendBuffer(self, byte):
		"""
		Appends the hex representation of a byte to the buffer
		"""
		self.dumpbuffer = self.dumpbuffer + ('%02X ' % byte)
		
		
	def write(self, byte):
		"""
		Writes a raw byte to parallel port.
		"""
		self.p.write(byte)
		
	
	def dump(self):
		"""
		Writes the buffer content to the dump file, then resets the buffer
		"""
		if self.dumpfile != None:
			try:
				self.dumpfile.write(self.dumpbuffer + '\n')
				self.dumpfile.flush()
				self.dumpbuffer = ''
			except Exception,e:
				pass
		
		